library(nhdplusTools)
library(dplyr)
library(sf)
library(hyRefactor)

# Load source data customized for HI
if(vpu == "20"){
  # Customizations for Hawaii
  crs <- 26904
  
  # Defined staged NHDPlus for HI
  staged_nhdplus_islands <- stage_national_data(nhdplus_data = data_paths$islands_gdb,
                                                output_path  = data_paths$islands_dir)
  
  # Hawaii source data created during 01_HI_Prep.Rmd
  raw_wbd <- "HI_WBD"
  gage_loc <- "GageLoc_20"
  gage_events_layer <- "gage_events"
  TE <- "TE_20"
  NID_layer <- "NID_20"
  WBs_layer <- "WB_20"
  data_paths$fullcats_table <- file.path(data_paths$islands_dir, "nhdcat_full.rds")
  
  # Refactor customizations for HI
  min_da_km <- 5
  split_meters <- 2500
  combine_meters <- 500
  elev_diff <- 300
  max_elev_DA <- 1.5
} else {
  # Cusomizations for AK
  crs <- 3338
  
  # Hawaii source data created during 01_HI_Prep.Rmd
  src_gpkg <- "cache/ak.gpkg"
  
  nav_gpkg <- file.path("cache", "reference_19.gpkg")
  
  # Data from source data
  ak_cats_layer <- "ak_cats"
  poi_layer <- paste0("POIS_", vpu)
  nhd_flowline <- "flowline"
  nhd_catchment <- "catchment"
  nhd_outlet <- "outlets"
  outlets_layer <- "hy_outlets"
  divide_layer <- "divides"
  events_DF_layer <- paste0("event_", vpu)
  nsegments_layer <- "nsegments"
  
  refactored_layer <- "collapsed"
  reconciled_layer <- "reconciled"
  agg_fline_layer <- "agg_fline"
  agg_cats_layer <- "agg_cats"
  cache_split <- "temp/19a.rda"
  
  lookup_table_file <- "cache/19_lookup.csv"
  
  rpu_code <- "19a"
  split_meters <- 10000
  combine_meters <- 1000
  min_da_km <- 20
  elev_diff <- 500
  max_elev_DA <- 20

  # parallel factor for catchment reconciliation.
  para_reconcile <- 2
  para_split_flines <- 2

}

# All instances of calls to this are commented out?
# getM <- function(i, events, nhd){
#   ##########################################
#   # Gets measure of flowline for split events
#   # 
#   # Arguments:
#   #   i : ID iterator
#   #   events : (SF, data frame) events 
#   #   nhd : (SF, data frame) nhd-like flowlines
#   #
#   # Returns:
#   #   Events in the format needed for aggregate flowlines
#   ########################################
#   event <- filter(events, id == i)
#   segment_int <- filter(nhd, COMID == event$COMID) %>%
#     rLFT::addMValues(.)
#   
#   segment_Z <- data.frame(st_coordinates(segment_int$geometry))
#   
#   segment_wZ <- segment_Z %>%
#     mutate(NHD_M = (1 - (M / as.numeric(st_length(segment_int)))) * 100, 
#            id = row_number())
#   
#   segment_woZ <- st_as_sf(segment_wZ, coords = c("X", "Y"), crs = st_crs(segment_int))
#   
#   nearest <- segment_woZ[which.min(st_distance(event, segment_woZ)),]
#   
#   event$M <- nearest$NHD_M
#   
#   return(event)
#   
# }

ak_wb_POIs_single <- function(wb_ID, single_seg, segs_int_wb){
  ##########################################
  # Delineates waterbody inlet/outlets for singe waterbody/flowline pair
  # 
  # Arguments:
  #   wb_ID : (Character) Waterbody ID 
  #   single_seg : (SF, data frame) segment, whole bisecting waterbody
  #   segs_int_wb : (SF, data frame) intersected segment, waterbody
  #
  # Returns:
  #   Waterbody inlet and oulet POIs for waterbodies
  #   for which multiple flowlines intersect
  ########################################
  
  # Segment associated with waterbody ID
  waterSegs <- filter(single_seg, OBJECTID == wb_ID)
  segment_int <- filter(segs_int_wb, COMID %in% waterSegs$COMID,
                        OBJECTID == wb_ID)
  
  # New Segment
  seg_in_wZ <- data.frame(st_coordinates(segment_int$geom)) %>% 
    select(X, Y) %>% 
    mutate(Xint = as.integer(X)) %>% 
    left_join(data.frame(st_coordinates(segment_int$geometry)) %>% 
                mutate(Xint = as.integer(X)) %>%
                select(Xint, M),
              by = "Xint") %>%
    st_as_sf(., coords = c("X", "Y"), crs = st_crs(segment_int)) %>%
    summarise(do_union = FALSE) %>%
    st_cast("LINESTRING", group = TRUE) %>%
    rLFT::addMValues(.)
  
  if (segment_int$arbolate_sum != segment_int$LENGTHKM){
    in_out <- data.frame(st_coordinates(seg_in_wZ$geometry)) %>%
      slice(1, n()) %>%
      mutate(wbID = wb_ID, comID = segment_int$COMID)
  } else {
    in_out <- data.frame(st_coordinates(seg_in_wZ$geometry)) %>%
      slice(n()) %>%
      mutate(wbID = wb_ID, comID = segment_int$COMID)
  }
 
  return(in_out) 
}

ak_wb_POIs <- function(wb_ID, segs_int_wb){
  ##########################################
  # Delineates waterbody inlet/outlets for which there area
  #            multiple pre-existing segments per single waterbody
  # 
  # Arguments:
  #   wb_ID : (Character) Waterbody ID 
  #   segs_int_wb : (SF, data frame) segments which intersect waterbodies
  #
  # Returns:
  #   Waterbody inlet and oulet POIs for waterbodies
  #   for which multiple flowlines intersect
  ########################################
  
  # where mutliple flowlines (COMID) intersect waterbody
  mult_seg <- segs_int_wb %>%
    group_by(OBJECTID) %>%
    filter(length(unique(COMID)) > 1) %>%
    ungroup()

  #  Segments that intersec waterboy
  segment_int <- filter(segs_int_wb, COMID %in% filter(mult_seg, OBJECTID == wb_ID)$COMID,
                        OBJECTID == wb_ID)
  
  # Outflow SEgment
  seg_out <- filter(segment_int, Hydroseq == min(Hydroseq)) 
  
  # New Segment
  seg_out_wZ <- data.frame(st_coordinates(seg_out$geom)) %>% 
    select(X, Y) %>% 
    mutate(Xint = as.integer(X)) %>% 
    left_join(data.frame(st_coordinates(seg_out$geometry)) %>% 
                mutate(Xint = as.integer(X)) %>%
                select(Xint, M),
              by = "Xint") %>%
    st_as_sf(., coords = c("X", "Y"), crs = st_crs(seg_out)) %>%
    summarise(do_union = FALSE) %>%
    st_cast("LINESTRING", group = TRUE) %>%
    rLFT::addMValues(.) 
  
  wb_out_poi <- data.frame(st_coordinates(seg_out_wZ$geometry)) %>%
    slice(n()) %>%
    mutate(wbID = wb_ID, comID = seg_out$COMID, type = "out") #%>%
    #st_as_sf(coords = c("X", "Y"), crs = st_crs(seg_out_wZ))
  
  # Inflow SEgment
  seg_in <- filter(segment_int, !COMID %in% seg_out$COMID) %>%
    # filter out segs completely in waterbody
    filter(st_length(geom) %>% as.integer() != 
             st_length(geometry) %>% as.integer()) 
  
  inPOIs <- function(inCOM, seg_in, wb_ID){
    seg_in_com <- filter(seg_in, COMID == inCOM) %>%
      st_cast("MULTILINESTRING")
    
    # New Segment
    seg_in_wZ <- data.frame(st_coordinates(seg_in_com$geom)) %>% 
      select(X, Y) %>% 
      mutate(Xint = as.integer(X)) %>% 
      left_join(data.frame(st_coordinates(seg_in_com$geometry)) %>% 
                  mutate(Xint = as.integer(X)) %>%
                  select(Xint, M),
                by = "Xint") %>%
      st_as_sf(., coords = c("X", "Y"), crs = st_crs(seg_in_com)) %>%
      summarise(do_union = FALSE) %>%
      st_cast("LINESTRING", group = TRUE) %>%
      rLFT::addMValues(.)
    
    wb_in_poi <- data.frame(st_coordinates(seg_in_wZ$geometry)) %>%
      slice(1) %>%
      mutate(wbID = wb_ID, comID = inCOM, type = "in") #%>%
      #st_as_sf(coords = c("X", "Y"), crs = st_crs(seg_out_wZ))
    
    return(wb_in_poi)
  }
  
  wb_in_poi <- do.call("rbind", lapply(c(seg_in$COMID), inPOIs, seg_in, wb_ID)) %>%
    bind_rows(wb_out_poi)
  
  return(wb_in_poi)
}

# unused?
# # http://rstudio-pubs-static.s3.amazonaws.com/10685_1f7266d60db7432486517a111c76ac8b.html
# CreateSegments <- function(coords, length = 0, n.parts = 0) {
#   stopifnot((length > 0 || n.parts > 0))
#   # calculate total length line
#   total_length <- 0
#   for (i in 1:(nrow(coords) - 1)) {
#     d <- sqrt((coords[i, 1] - coords[i + 1, 1])^2 + (coords[i, 2] - coords[i + 
#                                                                              1, 2])^2)
#     total_length <- total_length + d
#   }
#   
#   # calculate stationing of segments
#   if (length > 0) {
#     stationing <- c(seq(from = 0, to = total_length, by = length), total_length)
#   } else {
#     stationing <- c(seq(from = 0, to = total_length, length.out = n.parts), 
#                     total_length)
#   }
#   
#   # calculate segments and store the in list
#   newlines <- list()
#   for (i in 1:(length(stationing) - 1)) {
#     newlines[[i]] <- CreateSegment(coords, stationing[i], stationing[i + 
#                                                                        1])
#   }
#   return(newlines)
# }
  